"""
基础工具
"""

import numpy as np

from chippymu.models import DrumType, WaveType
from numpy import ndarray


def note_to_freq(*, note: int) -> float:
    """
    将 MIDI 音符编号转换为频率 (Hz)

    - 参数
        - note: MIDI 音符编号

    - 返回
        - 频率 (Hz)
    """
    return 440 * 2 ** ((note - 69) / 12)


def basic_wave_gen(
    *,
    wave_type: WaveType,
    frequency: float,
    duration: float,
    amplitude: float = 1.0,
    duty_cycle: float = 0.5,
    sample_rate: int = 16000,
) -> ndarray:
    """
    生成指定类型的波形数组。

    - 参数:
        - wave_type: 波形的类型，可以是 "sine"、"square"、"triangle"、"sawtooth" 或 "noise"。
        - frequency: 波形的频率，单位为赫兹（Hz）。
        - duration: 波形的持续时间，单位为秒（s）。
        - amplitude: 波形的振幅，范围从 0 到 1。
        - duty_cycle:  Square 波形的占空比，范围从 0 到 1。
        - sample_rate: 采样率，单位为赫兹（Hz），默认为16000，使用时可能从Params中获取。

    - 返回:
        - 一个包含波形数据的 NumPy 数组。
    """
    t: ndarray = np.linspace(
        start=0, stop=duration, num=int(sample_rate * duration), endpoint=False
    )
    phase: ndarray = (t * frequency) % 1

    match wave_type:
        case WaveType.SINE:
            return amplitude * np.sin(2 * np.pi * frequency * t)
        case WaveType.SQUARE:
            return amplitude * np.where(phase < duty_cycle, 1, -1)
        case WaveType.TRIANGLE:
            return amplitude * np.where(phase < 0.5, -1 + 4 * phase, 3 - 4 * phase)
        case WaveType.SAWTOOTH:
            return amplitude * (2 * phase - 1)
        # 将噪声移至鼓
        # case WaveType.NOISE:
        #     return amplitude * (2 * np.random.rand(len(t)) - 1)
        case _:
            raise ValueError("未知波形类型")


def generate_kick(
    *,
    duration: float,
    amplitude: float = 1.0,
    sample_rate: int = 16000,
) -> ndarray:
    """
    生成底鼓声音，低频正弦波带衰减。

    - 参数：
        - duration: 持续时间，单位秒
        - frequency: 频率，单位赫兹
        - amplitude: 振幅，范围0~1
        - sample_rate: 采样率，默认16000

    - 返回：
        - 音频数据，ndarray
    """
    frequency: int = 50
    t: ndarray = np.linspace(0, duration, int(sample_rate * duration), endpoint=False)
    envelope: ndarray = np.exp(-10 * t)  # 衰减
    return amplitude * envelope * np.sin(2 * np.pi * frequency * t)


def generate_snare(
    *, duration: float, amplitude: float = 1.0, sample_rate: int = 16000
) -> ndarray:
    """
    生成军鼓声音，噪声带衰减。

    - 参数：
        - duration: 持续时间，单位秒
        - amplitude: 振幅，范围0~1
        - sample_rate: 采样率，默认16000

    - 返回：
        - 音频数据，ndarray
    """
    t: ndarray = np.linspace(0, duration, int(sample_rate * duration), endpoint=False)
    noise: ndarray = 2 * np.random.rand(len(t)) - 1
    envelope: ndarray = np.exp(-20 * t)
    return amplitude * envelope * noise


def generate_hihat(
    *, duration: float, amplitude: float = 1.0, sample_rate: int = 16000
) -> ndarray:
    """
    生成踩镲声音，短噪声带快速衰减。

    - 参数：
        - duration: 持续时间，单位秒
        - amplitude: 音量
        - sample_rate: 采样率，默认16000

    - 返回：
        - 噪声, ndarray
    """
    t: ndarray = np.linspace(0, duration, int(sample_rate * duration), endpoint=False)
    noise: ndarray = 2 * np.random.rand(len(t)) - 1
    envelope: ndarray = np.exp(-50 * t)
    return amplitude * envelope * noise


def generate_noice(
    *, duration: float, amplitude: float = 1.0, sample_rate: int = 16000
) -> ndarray:
    """
    生成噪声，用于生成鼓声。

    - 参数：
        - duration: 持续时间，单位秒
        - amplitude: 振幅，范围0~1
        - sample_rate: 采样率，默认16000

    - 返回：
        - 噪声，ndarray
    """
    t: ndarray = np.linspace(0, duration, int(sample_rate * duration), endpoint=False)
    return amplitude * (2 * np.random.rand(len(t)) - 1)


def basic_drum_gen(
    *,
    drum_type: DrumType,
    frequency: int = 50,
    duration: float,
    amplitude: float = 1.0,
    sample_rate: int = 16000,
) -> ndarray:
    """
    生成基本鼓声

    - 参数：
        - drum_type: 鼓声类型
        - frequency: 频率，单位赫兹
        - duration: 持续时间，单位秒
        - amplitude: 振幅，范围0~1
        - sample_rate: 采样率，单位赫兹

    - 返回：
        - 音频数据，ndarray
    """
    match drum_type:
        case DrumType.KICK:
            return generate_kick(
                duration=duration,
                amplitude=amplitude,
                sample_rate=sample_rate,
            )
        case DrumType.SNARE:
            return generate_snare(
                duration=duration, amplitude=amplitude, sample_rate=sample_rate
            )
        case DrumType.HIHAT:
            return generate_hihat(
                duration=duration, amplitude=amplitude, sample_rate=sample_rate
            )
        case DrumType.NOICE:
            return generate_noice(
                duration=duration, amplitude=amplitude, sample_rate=sample_rate
            )
        case _:
            raise ValueError("未知鼓类型")



"""
按通道生成音G
"""

import numpy as np

from chippymu.utils import basic_wave_gen, basic_drum_gen, note_to_freq

from numpy import ndarray
from chippymu.configs import BasicParams
from chippymu.models import DrumType, Note, WaveType


def generate_wave(
    *, melody: list[tuple[float, int | Note, float]], wave_type: WaveType, params: BasicParams
) -> ndarray:
    """
    生成指定波形和旋律的通道

    - 参数：
        - melody: 旋律列表，每个元素为 (frequency, duration, amplitude)
        - wave_type: 波形类型
        - params: 基本参数

    - 返回：
        - 通道数据
    """
    melody_beat: list[tuple[float, int | Note, float]] = [
        (start * params.beat_duration, note, dur * params.beat_duration)
        for start, note, dur in melody
    ]

    melody_audio = np.zeros(params.whole_duration)

    for start, note, dur in melody_beat:
        wave = basic_wave_gen(
            wave_type=wave_type,
            frequency=note_to_freq(note=note),
            duration=dur,
            sample_rate=params.sample_rate,
        )
        start_idx = int(start * params.sample_rate)
        end_idx = start_idx + len(wave)
        melody_audio[start_idx:end_idx] += wave

    return melody_audio


def generate_drums(
    *, drums: list[tuple[float, DrumType, float]], params: BasicParams
) -> ndarray:
    """
    生成鼓声

    - 参数：
        - drums: 鼓声列表，每个元素是一个元组，包含起始时间，鼓声类型，持续时间
        - params: 基础参数

    - 返回：
        - 通道数据
    """
    drums_beat: list[tuple[float, DrumType, float]] = [
        (start * params.beat_duration, note, dur * params.beat_duration)
        for start, note, dur in drums
    ]

    drums_audio = np.zeros(params.whole_duration)

    for start, drum_type, dur in drums_beat:
        wave = basic_drum_gen(
            drum_type=drum_type,
            duration=dur,
            amplitude=1.0,
            sample_rate=params.sample_rate,
        )
        start_idx = int(start * params.sample_rate)
        end_idx = start_idx + len(wave)
        drums_audio[start_idx:end_idx] += wave

    return drums_audio



"""
定义一些基础类型
"""

from enum import Enum


class WaveType(str, Enum):
    """
    几种波形
    """

    SINE = "sine"
    SQUARE = "square"
    TRIANGLE = "triangle"
    SAWTOOTH = "sawtooth"
    # 移至鼓
    # NOISE = "noise"


class DrumType(str, Enum):
    """
    几种鼓声
    """

    KICK = "kick"
    SNARE = "snare"
    HIHAT = "hihat"
    NOICE = "noice"


class Note(int, Enum):
    """
    音符
    """

    C0 = 48
    C0_SHARP = 49
    D0 = 50
    D0_SHARP = 51
    E0 = 52
    F0 = 53
    F0_SHARP = 54
    G0 = 55
    G0_SHARP = 56
    A0 = 57
    A0_SHARP = 58
    B0 = 59
    
    C1 = 60
    C1_SHARP = 49
    D1 = 62
    D1_SHARP = 51
    E1 = 64
    F1 = 65
    F1_SHARP = 54
    G1 = 67
    G1_SHARP = 56
    A1 = 69
    A1_SHARP = 58
    B1 = 71

    C2 = 72
    C2_SHARP = 49
    D2 = 74
    D2_SHARP = 51
    E2 = 76
    F2 = 77
    F2_SHARP = 54
    G2 = 79
    G2_SHARP = 56
    A2 = 81
    A2_SHARP = 58
    B2 = 83

    C3 = 84
    C3_SHARP = 49
    D3 = 86
    D3_SHARP = 51
    E3 = 88
    F3 = 89
    F3_SHARP = 54
    G3 = 91
    G3_SHARP = 56
    A3 = 93
    A3_SHARP = 58
    B3 = 95

    C4 = 96
    C4_SHARP = 49
    D4 = 98
    D4_SHARP = 51
    E4 = 100
    F4 = 101
    F4_SHARP = 54
    G4 = 103
    G4_SHARP = 56
    A4 = 105
    A4_SHARP = 58
    B4 = 107


"""
后处理和音频播放
"""

import sounddevice as sd # type: ignore
import numpy as np

from numpy import ndarray

from chippymu.configs import BasicParams


def post_processing(channels: list[ndarray], volumes: list[float]) -> ndarray:
    """
    后处理过程，包括音量控制、混音、裁剪、量化
    """
    if len(channels) != len(volumes):
        raise ValueError("channels和volumes长度不一致")

    mix = np.zeros_like(channels[0])
    for i in range(len(channels)):
        mix += channels[i] * volumes[i]

    mix = np.clip(mix, -1, 1)

    mix_8bit = np.round(mix * 127).astype(np.int8)

    return mix_8bit


def play(data: ndarray, params: BasicParams):
    sd.play(data=data, samplerate=params.sample_rate)
    sd.wait()
    


"""
基础参数的配置文件
"""


class BasicParams:
    def __init__(self, *, sample_rate: int = 16000, bpm: int = 90, length: float):
        # 定义采样率
        self.sample_rate: int = sample_rate
        # 定义BPM
        self.bpm: int = bpm
        # 定义音频总拍数
        self.length: float = length

    @property
    def beat_duration(self) -> float:
        """
        计算每拍的时长
        """
        return 60.0 / self.bpm

    @property
    def whole_duration(self) -> int:
        """
        计算音频矩阵总长
        """
        return int(self.length * self.beat_duration * self.sample_rate)
        
